#!/bin/bash

if [[ ! "${1}" =~ ^\/Volumes\/. ]]; then
	echo "ERROR: This script must be executed with the /Volumes path of your SSD (such as \"./verifytrim.sh /Volumes/Samsung\"). Also be sure that your SSD is empty before you run this test, and that TRIM is enabled in OS X."
	exit 1
fi
if [[ ! -d "${1}" ]]; then
	echo "ERROR: '${1}' is not a valid volume name!"
	exit 1
fi
echo "- Entering ${1}..."
cd "${1}"

# Hello!
echo "[[ * ]] This script verifies that TRIM is working properly on your SSD. Some modern SSDs, and especially older ones made <= 2012, will randomly corrupt data when TRIM is executed. We test your SSDs TRIM implementation by creating a 50 GB file with a known content, and then TRIM'ing the rest of the drive's free space. If the TRIM is broken, it will break data inside the valid regions of the 50 GB file, and we will detect this corruption when we verify the hash at the end of this script. The whole analysis process takes about an hour, and requires that your disk is empty and that it will be left idle to do its garbage collection. More information will come in the subsequent prompts..."
sleep 5

# Step 1: Generate a 50 GB trimtest.tmp filled with binary 1's.
if [ ! -f trimtest.tmp ]; then
	echo "* Creating trimtest.tmp and filling it with 50 GB of binary 1's. The data is generated by perl, which is a bit slow, and will therefore probably be written at around 100-130 MB/second. At that rate, it will take about 8.5 minutes. You can check the progress by opening Activity Monitor, switching to the Disk tab and searching for 'dd' to see how much data it has written so far."
	perl -e '$gigabytes=50; do{ $kbOnes="$kbOnes\xFF"; }while( length($kbOnes)<1024 ); $mbSize=1024*1024; do{ $mbOnes="$mbOnes$kbOnes"; }while( length($mbOnes)<$mbSize ); for( $megabytes=1, $maxmegabytes=$gigabytes*1024; $megabytes<=$maxmegabytes; ++$megabytes ) { print $mbOnes; }' | dd of=trimtest.tmp
else
	echo "* Your trimtest.tmp file already exists, so we will use that one."
fi
echo "* Verifying that trimtest.tmp is exactly 50 GB."
TMPSIZE=$(du -k trimtest.tmp | cut -f1)
if [ "${TMPSIZE}" -ne 52428800 ]; then
	echo "* Your trimtest.tmp is the wrong size. Please delete it and re-run this script."
	exit 1
else
	echo "* Size is okay. Proceeding..."
fi

# Step 2: Fill the rest of the disk space with zeroes.
SSDFREE=$(df -m "${PWD}" | tail -1 | sed -E 's/ {2,}/ /g' | cut -d' ' -f4)
SSDFILL=$(( $SSDFREE - 1024 ))
if [ "${SSDFILL}" -le 0 ]; then SSDFILL=1; fi
SSDFILLGBS=$(awk -v mbs=$SSDFILL 'BEGIN { rounded = sprintf("%.0f", mbs/1024); print rounded }')
echo "* Your SSD has ${SSDFREE} megabytes of free space. We will now fill up ${SSDFILL} megabytes (~${SSDFILLGBS} gigabytes), leaving only 1 GB of free space at the end. This ensures thorough coverage of TRIM across your entire drive. The write is performed using a simple zero-fill which means that it should be able to max out your drive's sustained write-speed, which is usually around 480-550 MB/second for good drives. The time depends on how much data will be written and the speed of your drive, so no time estimate can be given, but for reference I filled 425 gigabytes in 15 minutes on my Samsung 850 Pro. You can check the progress by opening Activity Monitor, switching to the Disk tab and searching for 'dd' to see how much data it has written so far."
rm -f zerofill.tmp
dd if=/dev/zero of=zerofill.tmp bs=1m count=${SSDFILL}

# Step 3: The user must now manually delete the zerofill.tmp file via the Finder.
while :; do
	echo "* You must now MANUALLY open the Finder and navigate to your SSD, and then delete zerofill.tmp. That will trigger a TRIM for the ~${SSDFILLGBS} gigabytes consumed by the file. Type YES when you are done and ready to continue."
	read -r -p "- Have you deleted zerofill.tmp via the Finder? [yes/N] " response
	if [[ $response =~ ^[yY][eE][sS]$ ]]; then
		if [ -f zerofill.tmp ]; then
			echo "* NOPE! ZEROFILL.TMP STILL EXISTS! GO DELETE IT NOW AND READ THESE INSTRUCTIONS MORE CAREFULLY!"
		else
			# good, they've deleted the file, so break the infinite loop!
			break
		fi
	fi
done

# Step 4: Sleep for half an hour to let the SSD's TRIM/Garbage Collection run in the background.
echo "* IMPORTANT: You must now let the drive REST for half an hour so that the internal garbage collection will erase the blocks that have been trimmed. This script will now sleep for 30 minutes, and will be counting down each minute interval to let you know when it's getting close to resuming..."
for MINUTESLEFT in {30..1}; do
	echo "- ${MINUTESLEFT} minutes left before resuming..."
	sleep 60
done

# Step 5: Verify the SHA1 hash of the file.
echo "* Checking the SHA1 hash of trimtest.tmp to see if your SSD has corrupted any valid data on the disk. The correct hash should be 'c52e363998efdff714c01bbbf6be4ba21a5fffc0'. This will probably take about 5 minutes. You can check the progress by opening Activity Monitor, switching to the Disk tab and searching for 'openssl' to see how much data it has read so far."
TRIMHASH=$(openssl sha1 trimtest.tmp | cut -d' ' -f2)
if [ "${TRIMHASH}" = "c52e363998efdff714c01bbbf6be4ba21a5fffc0" ]; then
	echo "* Hash okay. Your SSD has a properly working TRIM implementation! Congratulations!"
else
	echo "* Hash mismatch. The trimtest.tmp file is damaged on disk. Your SSD has a broken TRIM implementation which corrupts data."
fi